package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.espec.ESpecLogic;
import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.parser.ESpecDefaults;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import be.iminds.ilabt.jfed.espec.util.LimitedLiveLog;
import be.iminds.ilabt.jfed.espec.util.LimitedLiveLogImpl;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.highlevel.util.InputStreamToLogsThread;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.connection.channel.direct.Session;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.event.Level;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.util.*;

/**
 * ExperimentSpecificationServicesDirsAndUploadsState executes a
 * single execute statement in an ExperimentSpecification, on all nodes (that are specified).
 * It waits until the execute is completed before being finished
 */
public class ExperimentSpecificationServicesSingleExecuteState extends SlicedState<ExperimentSpecificationServicesSingleExecuteState.ExperimentSpecificationServicesSingleExecuteStateSlice> {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentSpecificationServicesSingleExecuteState.class);
    public static final int MAX_EXECUTE_FINISHED_FILE_DETECTION_TRIES = 60;
    public static final int TIME_BETWEEN_FINISHED_FILE_DETECTION_TRIES_MS = 10_000;

    @Nonnull private final SetupSoftwareExperimentJob job;
    @Nonnull private final Experiment experiment;

    @Nonnull private final ExperimentSpecification experimentSpecification;
    @Nonnull private final ESpecBundle experimentSpecificationBundle;
    @Nonnull private final ESpecLogger eSpecLogger;

    @Nonnull private final List<ExperimentSpecificationServicesSingleExecuteStateSlice> slices;

    @Nonnull private final ExperimentSpecificationServicesDirsAndUploadsState parent;
    @Nonnull private final ExecuteSpec executeSpec;


    protected ExperimentSpecificationServicesSingleExecuteState(@Nonnull SetupSoftwareExperimentJob job,
                                                                @Nonnull ExperimentSpecificationServicesDirsAndUploadsState parent,
                                                                @Nonnull ExecuteSpec executeSpec,
                                                                boolean isInAnsible) {
        super("ExperimentSpecification execute "+ executeSpec.getDesc());
        this.job = job;
        this.experiment = job.getExperiment();
        this.eSpecLogger = job.getExperiment().getExperimentSpecificationLogger();
        this.parent = parent;
        this.executeSpec = executeSpec;

        if (experiment.getExperimentSpecification() == null)
            throw new RuntimeException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification");
        if (experiment.getExperimentSpecificationBundle() == null)
            throw new RuntimeException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification (bundle)");

        this.experimentSpecification = experiment.getExperimentSpecification();
        this.experimentSpecificationBundle = experiment.getExperimentSpecificationBundle();

        experiment.requireSlice();

        this.slices = new ArrayList<>();

        assert !isInAnsible || experimentSpecification.getAnsible() != null;

        RequestRspecSource requestRspecSource = experiment.getSlice().getRequestRspec();
        FXModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec(ModelRspecType.FX);
        assert model != null;

        if (model == null) {
            LOG.warn("ExperimentSpecificationServicesSingleExecuteState will do nothing because there is no usable request RSpec");
            return;
        }

        for (FXRspecNode node : model.getNodes()) {
            String nodeId = node.getClientId();

            if (nodeId != null) {
                if (ESpecLogic.mustRunOn(
                        nodeId, node.getAnsibleGroups(),
                        experimentSpecification, executeSpec)) {
                    slices.add(new ExperimentSpecificationServicesSingleExecuteStateSlice(job, node));
                }
            }
        }
    }

    @Nonnull
    @Override
    protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
        ExperimentTaskStatus res = super.executeState(job);
        experiment.getExperimentSpecificationLogger().firePostExecuteStepAll(executeSpec, res == ExperimentTaskStatus.SUCCESS);
        return res;
    }

    @Override
    public Collection<ExperimentSpecificationServicesSingleExecuteStateSlice> getSlices() {
        return slices;
    }

    public class ExperimentSpecificationServicesSingleExecuteStateSlice extends StateSlice {
        @Nonnull protected final SetupSoftwareExperimentJob setupSoftwareExperimentJob;
        @Nonnull protected final FXRspecNode node;
        @Nonnull protected final String clientId;

        public ExperimentSpecificationServicesSingleExecuteStateSlice(@Nonnull SetupSoftwareExperimentJob job,
                                                                      @Nonnull FXRspecNode node) {
            super(job);
            this.setupSoftwareExperimentJob = job;
            this.node = node;

            if (node.getClientId() == null)
                throw new IllegalStateException("Only nodes with a client_id can be used in an ExperimentSpecification");
            this.clientId = node.getClientId();
        }

        @Override
        public String getName() {
            return "ExperimentSpecification execute steps for node "+node.getClientId();
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            SSHClient ssh = null;
            try {
                ssh = setupSoftwareExperimentJob.getSSHClient(node, this);
            } catch (SetupSoftwareExperimentJob.SshException e) {
                throw new IllegalStateException("There is no SSH connection for node \""+clientId+"\" available", e);
            }

            if (!ssh.isConnected()) {
                updateMessage("SSH connection is not connection: will abort execute.");
                return ExperimentTaskStatus.FAILED;
            }

            final ExecuteState executeState = new ExecuteState(ssh);
            setAndRunState(executeState);

            if (executeState.getStatus() != ExperimentTaskStatus.SUCCESS)
                return ExperimentTaskStatus.FAILED;
//
//            //TODO: later, we might need something like this for ADVANCED scripts
////            //check if there are finished-files to wait for
////            if (node.getExecuteServices().stream().anyMatch(es -> es.getFinishedFile() != null)) {
////                setAndRunState(new WaitForExecuteServicesState());
////            } else {
////                LOG.info("Did not find any execute-services with finished-file attribute. We assume that all files that are needed are present!");
////            }

            return ExperimentTaskStatus.SUCCESS;
        }

//        public class WaitForExecuteServicesState extends State {
//
//            protected WaitForExecuteServicesState() {
//                super("Waiting for execute-services to finish on " + node.getClientId());
//            }
//
//            @Nonnull
//            @Override
//            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
//                try {
//                    for (FXExecuteService executeService : node.getExecuteServices()) {
//                        if (executeService.getFinishedFile() == null)
//                            continue;
//
//                        boolean fileFound = false;
//                        int tries = 0;
//
//                        while (!fileFound && tries < MAX_EXECUTE_FINISHED_FILE_DETECTION_TRIES) {
//                            Session.Command command =
//                                    ssh.startSession().exec("test -f " + executeService.getFinishedFile());
//                            command.join();
//
//                            fileFound = command.getExitStatus() == 0;
//
//                            tries++;
//                            if (!fileFound) {
//                                updateMessage("Looked for '" + executeService.getFinishedFile() + "' " + tries + " times.");
//                                Thread.sleep(TIME_BETWEEN_FINISHED_FILE_DETECTION_TRIES_MS);
//                            }
//                        }
//
//                        if (!fileFound) {
//                            updateMessage("Timed out while waiting for finished-file '" + executeService.getFinishedFile() + "' to appear");
//                            return ExperimentTaskStatus.FAILED;
//                        } else {
//                            updateMessage("Found finished-file '" + executeService.getFinishedFile() + "'");
//                        }
//                    }
//
//                    updateMessage("Found all finished-files");
//                    return ExperimentTaskStatus.SUCCESS;
//
//
//                } catch (IOException ex) {
//                    LOG.error("Error while waiting for finished-files on {}", node.getClientId(), ex);
//                    updateMessage("Error while waiting for finished-files on {}" + ex.getMessage());
//                    return ExperimentTaskStatus.FAILED;
//                }
//            }
//        }

        public class ExecuteState extends State {
            @Nonnull private final SSHClient ssh;

            protected ExecuteState(@Nonnull SSHClient ssh) {
                super("Executing "+executeSpec.getDesc()+" on " + node.getClientId());
                this.ssh = ssh;
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                if (ssh == null)
                    throw new IllegalStateException("ssh == null");

                String desc = executeSpec.getDesc();
                long logId = eSpecLogger.getNextLogId();

                if (!ESpecLogic.mustRunOn(
                        clientId, node.getAnsibleGroups(),
                        experimentSpecification, executeSpec)) {
                    throw new IllegalStateException("BUG: execute "+desc+" does not need to be executed on this " +
                            "node (\""+clientId+"\") (nodes="+executeSpec.getNodes()+")");
                }

                String remotePath = parent.getRemotePath(executeSpec, clientId);
                if (remotePath == null) {
                    eSpecLogger.firePostExecuteFail(logId, executeSpec, "not found", null, clientId, -1,
                            "Did not find remote path of executable. Bug in execute step?", null);
                    LOG.error("Did not find remote path of executable. Bug in execute step?");
                    updateMessage("Did not find remote path of executable. Bug in execute step?");
                    throw new RuntimeException("Did not find remote path of executable. Bug in execute step?");
//                    return ExperimentTaskStatus.FAILED;
                }

                SshCommandAndLog commandAndLog = createSshCommand(remotePath,
                        executeSpec, experimentSpecification, parent, clientId,
                        executeSpec.getPwd(),
                        executeSpec.isSudo());

                try {
                    LOG.debug("ExperimentSpecification.execute is executing {} on {}",
                            remotePath, node.getClientId());
                    LOG.debug("      command: {}", commandAndLog.getCommand());


                    LimitedLiveLog log = new LimitedLiveLogImpl(1_000_000);
                    eSpecLogger.firePreExecute(logId, executeSpec, remotePath,
                            commandAndLog.getLogPath(), clientId, log);

                    Session.Command c =
                            ssh.startSession().exec(commandAndLog.getCommand());

                    new InputStreamToLogsThread(c.getInputStream(), LOG, Level.INFO,
                            desc+"-OUTPUT: ", log).start();
                    new InputStreamToLogsThread(c.getErrorStream(), LOG, Level.WARN,
                            desc+"-ERROR: ", log).start();

                    c.join();

                    int exitStatus = c.getExitStatus();

                    if (exitStatus != 0) {
                        eSpecLogger.firePostExecuteFail(logId, executeSpec, remotePath, commandAndLog.getLogPath(), clientId, exitStatus, desc+
                                " execution was unsuccessful. Aborting", null);
                        updateMessage(desc+" execution was unsuccessful. Aborting");
                        return ExperimentTaskStatus.FAILED;
                    } else {
                        eSpecLogger.firePostExecuteSuccess(logId, executeSpec, remotePath, commandAndLog.getLogPath(), clientId, exitStatus);
                        return ExperimentTaskStatus.SUCCESS;
                    }
                } catch (IOException ex) {
                    eSpecLogger.firePostExecuteFail(logId, executeSpec, remotePath, commandAndLog.getLogPath(), clientId, -1,
                            "Error while executing "+remotePath+" on "+node.getClientId(), ex);
                    LOG.error("Error while executing {} on {}", remotePath, node.getClientId(), ex);
                    updateMessage(String.format("Error while executing %s on %s: %s", remotePath, node.getClientId(), ex.getMessage()));
                    return ExperimentTaskStatus.FAILED;
                }
            }
        }
    }

    private static class SshCommandAndLog {
        @Nonnull private final String command;
        @Nullable private final String logPath;

        public SshCommandAndLog(@Nonnull String command, @Nullable String logPath) {
            this.command = command;
            this.logPath = logPath;
        }

        @Nonnull
        public String getCommand() {
            return command;
        }

        @Nullable
        public String getLogPath() {
            return logPath;
        }
    }
    public static SshCommandAndLog createSshCommand(@Nonnull String remotePath,
                                                        @Nonnull UploadLikeWithLogSpec executeSpec,
                                                        @Nonnull ExperimentSpecification experimentSpecification,
                                                        @Nonnull ExperimentSpecificationServicesDirsAndUploadsState parent,
                                                        @Nonnull String clientId,
                                                        @Nullable String pwd,
                                                        @Nullable Boolean useSudo) {
        String logPath;
        if (executeSpec.getLog() == null) {
            if (experimentSpecification.getConfig().isStoreRemoteLogsByDefault())
                logPath = ESpecDefaults.deriveAutomaticLogFile(remotePath);
            else
                logPath = null;
        } else {
            logPath = parent.getRemoteLogPath(executeSpec, clientId);
            assert logPath != null :
                    "Bug when resolving log path \""+executeSpec.getLog()+"\" for "+executeSpec.getDesc();
        }

        if (pwd == null) {
            //use working dir of script
            pwd = ExternalFileUtil.getPathOnly(remotePath);
        }

        return new SshCommandAndLog(
                createSshCommand(remotePath, pwd, logPath, useSudo, experimentSpecification),
                logPath
        );
    }

    public static String createSshCommand(@Nonnull String baseCommand,
                                          @Nullable String pwd,
                                          @Nullable String logPath,
                                          @Nullable Boolean useSudo,
                                          @Nonnull ExperimentSpecification experimentSpecification) {
        boolean localLogs = experimentSpecification.getConfig().isStoreLocalLogsByDefault();

        String sudoargs = "";
        if (experimentSpecification.getConfig().getDefaultSudoUser() != null) {
            sudoargs += "-u "+experimentSpecification.getConfig().getDefaultSudoUser()+" ";
        }

        if (useSudo == null) {
            useSudo = experimentSpecification.getConfig().isUseSudoByDefault();
        }

        String hereDocCommand = "<<'EOF'";

        String command;
        if (useSudo) {
            if (logPath != null) {
                if (localLogs)
                    command = "set -o pipefail; sudo " + sudoargs + "bash " + hereDocCommand + " 2>&1 | sudo tee " + logPath + "";
                else
                    command = "sudo " + sudoargs+"bash " + hereDocCommand + " > " + logPath + " 2>&1'";
            } else {
                if (localLogs)
                    command = "sudo " + sudoargs + hereDocCommand;
                else
                    command = "sudo " + sudoargs + hereDocCommand+" > /dev/null 2>&1";
            }
        } else {
            if (logPath != null) {
                if (localLogs)
                    command = "set -o pipefail; bash "+ hereDocCommand + " 2>&1 | tee '" + logPath + "'";
                else
                    command = "bash "+hereDocCommand + " > '" + logPath + "' 2>&1";
            } else {
                if (localLogs) {
//                    return "bash " + baseCommand;
                    command = "bash "+hereDocCommand+"";
                } else
                    command = "bash "+hereDocCommand+" > /dev/null 2>&1";
            }
        }
        command += "\n";
        if (pwd != null) {
            command += "cd '"+pwd+"'\n";
        }
        command += baseCommand+"\n";
        command += "EOF\n";
        return command;
    }

}
